<?php 
session_start();
// Turn off errors and warning to avoid messing up the API output
//ini_set('display_errors', 1);
//ini_set('display_startup_errors', 1);
//error_reporting(E_ALL);
error_reporting(0);
// TaskQueue class, PHP side
class TaskQueue {
  // Path to the SQLite3 databse on the server
  private $pathDb;
  // SQLite3 database object
  private $db;
  // Max number of workers
  private $nbMaxWorker;
  // Database name obfuscation parameter
  private $dbObfuscation;
  // Database definition
  private $dbDef;
  // Constructor
  function __construct() {
    // Obfuscation parameter for the database name, set this parameter to a strong
    // password before uploading this web app to a web server
    $this->dbObfuscation = "";
    // Check that the obfuscation parametr is set
    if($this->dbObfuscation == "") {
      echo "The obfuscation parameter for the database hasn't been set. " .
        "Set it before using TaskQueue.";
      exit(0);
    }
    // Set the path to the database, it is located in the same folder as the
    // php script, this folder must have reading and writing permission for
    // the account executing the script
    $this->pathDb = dirname(__FILE__) .
      "/TaskQueue." . $this->dbObfuscation . ".db";
    // Database model definition
    $this->dbDef = array(
      'TaskConfig' => array(
        'NbMaxWorker' => 'INTEGER'),
      'Task' => array(
        'Key' => 'TEXT',
        'Status' => 'INTEGER', // 0: pending, 1: running, 2: complete, 3: failed
        'Label' => 'TEXT',
        'Owner' => 'TEXT',
        'WorkerMsg' => 'TEXT',
        'Input' => 'INTEGER',
        'Output' => 'INTEGER')
    );
    // Try to open the database, on first access it creates it
    try {
      $this->db = new SQLite3($this->pathDb);
    } catch(Exception $e) {
      echo "<div class='divTile'>Exception caught:<br>" . nl2br($e) . "</div>";
    }
    // Set the max number of workers
    $this->nbMaxWorker = 10;
  }
  // Destructor
  function __destruct() {
    // Close the database connection if it's opened
    if($this->db) $this->db->close();
  }
  // Main method
  function Main() {
    try {
      // Check if the database is ready to use
      $this->CheckDb();
    } catch(Exception $e) {
      echo "<div class='divTile'>Exception caught:<br>" . nl2br($e) . "</div>";
    }
  }
  // Ensure the database is ready to use
  function CheckDb() {
    // Loop on the tables in the database definition
    foreach ($this->dbDef as $table => $tableDef) {
      // If the table doesn't exist, create it
      $sql = 'CREATE TABLE IF NOT EXISTS ' . $table;
      $sql .= ' (Reference INTEGER PRIMARY KEY';
      foreach ($tableDef as $field => $fieldDef) {
        $sql .= ', ' . $field . ' ' . $fieldDef;
      }
      $sql .= ')';
      $this->db->exec($sql);
    }
    // Set the maximum number of workers in the configuration
    $sql =
      'INSERT OR REPLACE INTO TaskConfig (Reference, NbMaxWorker) VALUES (1, ' .
      $this->nbMaxWorker . ')';
    $this->db->exec($sql);
  }
  // Entry point for the API
  function API() {
    try {
      // Route the API request to the appropriate method
      if(isset($_POST['action']) == false) {
        echo '{"actionResult":"no_action"}';
      } else if($_POST['action'] == 'getTasks') {
        $this->GetTasks();
      } else if($_POST['action'] == 'addTask') {
        $this->AddTask();
      } else if($_POST['action'] == 'deleteTask') {
        $this->DeleteTask();
      } else {
        echo '{"actionResult":"invalid_action"}';
      }
    } catch(Exception $e) {
      echo '{"actionResult":"exception_caught"}';
    }
  }
  // Process an API request for add a task
  function AddTask() {
    try {
      // Ensure there is no double quote in the labels
      $_POST['label'] = str_replace('"', ' ', $_POST["label"]);
      // Prepare the statement for addition of a new task
      $sql =
        "INSERT INTO Task " .
        "(Key, Status, Label, Owner, Input) " .
        "VALUES (:key, :status, :label, :owner, :input)";
      while(($stmt = $this->db->prepare($sql)) == false) usleep(10000);
      // Bind and execute the statement
      $stmt->bindValue(":key", "", SQLITE3_TEXT);
      $stmt->bindValue(":status", 0, SQLITE3_INTEGER);
      $stmt->bindValue(":label", $_POST["label"], SQLITE3_TEXT);
      $stmt->bindValue(":owner", $_POST["owner"], SQLITE3_TEXT);
      $stmt->bindValue(":input", $_POST["input"], SQLITE3_INTEGER);
      while($stmt->execute() == false) usleep(10000);
      // Spawn a worker (if possible)
      $log = $this->SpawnWorker();
      // Create the response
      echo '{"actionResult":"ok", "msg":"added task ' . $_POST["label"] . '", "log":"' . $log . '"}';
    } catch (Exception $e) {
      echo '{"actionResult":"error", "msg":"' . $e . '"}';
    }
  }
  // Process an API request for removal of a task
  function DeleteTask() {
    try {
      // Prepare the statement for removing
      $sql = "DELETE FROM Task WHERE Reference=:ref AND Owner=:owner";
      while(($stmt = $this->db->prepare($sql)) == false) usleep(10000);
      // Bind and execute the statement
      $stmt->bindValue(":ref", $_POST['ref'], SQLITE3_INTEGER);
      $stmt->bindValue(":owner", $_POST['owner'], SQLITE3_TEXT);
      while($stmt->execute() == false) usleep(10000);
      // Create the response
      echo '{"actionResult":"ok", "msg":"removed task ' . $_POST["label"] . '"}';
    } catch (Exception $e) {
      echo '{"actionResult":"error", "msg":"' . $e . '"}';
    }
  }
  // Process an API request to get the tasks of a user
  function GetTasks() {
    try {
      // Create the statement
      $sql = 
        "SELECT Reference, Key, Status, Label, Owner, WorkerMsg, " .
        "Input, Output FROM Task WHERE Owner=:owner " .
        "ORDER BY Label";
      while(($stmt = $this->db->prepare($sql)) == false) usleep(10000);
      // Bind and execute the statement
      $stmt->bindValue(":owner", $_POST["owner"], SQLITE3_TEXT);
      while(($rows = $stmt->execute()) == false) usleep(10000);
      // Convert the results into a JSON string
      $sep = ' ';
      $tasks = '';
      if($rows) while($row = $rows->fetchArray()) {
        $tasks .=
          $sep . '{' . 
          '"Reference":"' . $row['Reference'] . '",' .
          '"Key":"' . $row['Key'] . '",' .
          '"Status":"' . $row['Status'] . '",' .
          '"Label":"' . $row['Label'] . '",' .
          '"Owner":"' . $row['Owner'] . '",' .
          '"WorkerMsg":"' . $row['WorkerMsg'] . '",' .
          '"Input":"' . $row['Input'] . '",' .
          '"Output":"' . $row['Output'] . '"}';
        $sep = ', ';
      }
      // Create the response
      echo '{"actionResult":"ok", "msg":"", "tasks":[' . $tasks . ']}';
    } catch (Exception $e) {
      echo '{"actionResult":"error", "msg":"' . $e . '"}';
    }
  }
  // Spawn a new worker if possible
  function SpawnWorker() {
    try {
      // If there a not too many workers
      $rows = $this->db->query("SELECT COUNT(*) FROM Task WHERE Status=1");
      $nbRunning = $rows->fetchArray(SQLITE3_NUM)[0];
      if($nbRunning < $this->nbMaxWorker) {
        // Spawn a new worker
        $cmd = dirname(__FILE__) . '/worker ' .
          $this->dbObfuscation . ' > /dev/null &';
        $ret = exec($cmd);
      }
    } catch (Exception $e) {
      return str($e);
    } 
    return "";
  }
}
// Create the TaskQueue instance
$taskQueue = new TaskQueue();
// Process the eventual API request
if(isset($_POST['action'])) {
  $taskQueue->API();
  // When responding as an API, stop the script execution here to avoid the
  // DOM to be sent and mess up the API response
  exit(0);
// Else, call the main function
} else {
  $taskQueue->Main();
}
?>
<!DOCTYPE html>
<html>
  <head>
    <meta charset="UTF-8">
    <meta name="viewport" content="width=device-width,initial-scale=1">
    <style>
html { width: 100%; height: 100%; }
body {
  background-color: #aaaaaa; color: #433126;
  font-family:  serif; font-size: 20px;
  width: 100%; height: 100%; margin: 0;
}
#divFooter {
  text-align: center; font-size: 12px;
  margin: auto; margin-top: 20px; width: 100%; padding: 5px;
  background-color: #aaaaaa;
}
input[type="button"] {
  background-color: #fecb5e; box-shadow: 2px 2px 10px #888888;
  margin: 2px 5px; padding: 2px 4px; height: 24px;
  border: 1px solid #fee9aa; border-radius: 5px; color: #624838;
}
input[type="submit"] {
  background-color: #fecb5e; box-shadow: 2px 2px 10px #888888;
  margin: 2px 5px; padding: 2px 4px; height: 24px;
  border: 1px solid #fee9aa; border-radius: 5px; color: #624838;
}
select {
  background-color: #fecb5e; box-shadow: 2px 2px 10px #888888; color: #624838;
  margin: 2px 5px; padding: 2px 4px; max-width: 290px;
  font: 13px sans-serif; text-decoration: none; font: 13px sans-serif;
  border: 1px solid #fee9aa; border-radius: 5px;
}
#divMain { margin:auto; text-align: center; width: 90%;}
#divTitle { font-size: 30px; margin:auto; text-align: center; }
.divTile {
   width: 45%; min-width: 400px; vertical-align: top;
  position: relative; margin: 10px; padding: 10px;
  border: 1px solid #888888; border-radius: 3px; background-color: #cccccc;
  display: inline-block; overflow: auto; box-shadow: 4px 4px 4px 2px #999999;
}
#divLogin { margin:auto; }
.divInput { display: block; }
.divInputLeft { 
  width: 48%; text-align: right; display: inline-block; vertical-align: top;
}
.divInputRight { width: 48%; text-align: left; display: inline-block; }
.divTask {
  display: inline-block; width: 90%; padding:5px; margin: 1px;
  border: 1px solid #999999; border-radius: 3px; background-color: #dddddd;
  display: inline-block; overflow: auto; box-shadow: 2px 2px 10px #aaaaaa;
}
#divTasks { max-height: 500px; overflow: auto; }
    </style>
  </head>
  <title> TaskQueue </title>
  <body>
    <div id="divMain">
      <div id="divTitle" class="divRow">
        TaskQueue
      </div>
      <div id='divInput' class='divTile'>
        <div class='divInput'>
          <label class='divInputLeft'>Label:</label>
          <input type='text' id='inpLabel' class='divInputRight' placeholder='new task'>
        </div>
        <div class='divInput'>
          <label class='divInputLeft'>input:</label>
          <input type='number' id='inpInput' value='0' class='divInputRight'>
        </div>
        <input id='btnAddTask' type='button' value='Add'>
        <hr>
        <div id="divTasks">
        </div>
      </div>
    </div>
   </body>
  <script>
// Helper function to get an element by its id
function elem(id) { return document.getElementById(id); }
// TaskQueue class, Javascript side
class TaskQueue {
  constructor() {
    // A complete app would have login funcitonalities and this property
    // would be initialised accordingly
    this.login = "login"
    // Current tasks
    this.tasks = {};
  }
  // Method to create and add a new element
  AddElem(parent, type, properties) {
    let elem = document.createElement(type);
    for (const [prop, val] of Object.entries(properties)) elem[prop] = val;
    parent.appendChild(elem);
    return elem;
  }
  // Method to post an API request
  PostRequest(reqData, handler) {
    // Prepare the request data by adding the authentication data
    reqData.setAttribute("method", "post");
    let formData = new FormData(reqData);
    // Create the HTTP request
    let xhr = new XMLHttpRequest();
    xhr.onreadystatechange = function() {
      // If the HTTP request is complete and successful
      if (this.readyState == 4 && this.status == 200) {
        // Parse the response in JSON format
//console.log(this.responseText);
        let retData = JSON.parse(this.responseText);
        // If the token as expired
        if(retData["actionResult"] == "token_expired") {
          // Redirect the user to the login page
          window.location.href = 'index.php';
        // Else process the response with the handler in argument
        } else handler(retData);
      // Else, if the HTTP request is complete and not successful
      } else if (this.readyState == 4 && this.status != 200) {
        console.log('API request failed. ' + this.status);
      // Else, the HTTP request is not complete, wait for completion
      }
    };
    // Send the request
    xhr.open('POST', 'index.php');
    xhr.send(formData);
  }
  // Convert a dictionary of data for the API request into a form element
  // usable by the HTTP request
  DictToReqData(data) {
    let reqData = document.createElement("form");
    for (const [lbl, val] of Object.entries(data)) {
      this.AddElem(reqData, 'input', {'type': 'text', 'name':lbl, 'value':val});
    }
    return reqData;
  }
  // Add a new task to the queue
  AddTask() {
    // Avoid creating task without a label
    if(elem("inpLabel").value.trim() == "") {
      console.log("Can't create a task without a label");
    } else {
      // Create the API request data
      let data = {
        'action': 'addTask',
        'label': elem("inpLabel").value,
        'owner': this.login,
        'input': elem("inpInput").value,
      };
      let reqData = this.DictToReqData(data);
      // Post the API request
      this.PostRequest(reqData,
        // Handler for the API request response
        (retData) => {
          // Update the DOM
          this.UpdateTasks();
        });
    }
  }
  // Update the list of tasks
  UpdateTasks() {
    // Create the API request data
    let data = {
      'action': 'getTasks',
      'owner': this.login,
    };
    let reqData = this.DictToReqData(data);
    // Post the API request
    this.PostRequest(reqData,
      // Handler for the API request response
      (retData) => {
        // If the API requests was successful
        if(retData['actionResult'] == 'ok') {
          // Delete from the DOM the currently displayed tasks
          let divTasks = elem("divTasks");
          while(divTasks.childElementCount > 0)
            divTasks.removeChild(divTasks.firstElementChild);
          // For each task
          retData["tasks"].forEach(task => {
            // Add the task to the DOM
            let ref = task["Reference"];
            let divTask = this.AddElem(divTasks, 'div',
              {'className':'divTask', 'id':'divTask' + ref});
            let taskDesc = task["Label"] + ": " + task["Input"] + "→";
            if(task["Status"] == 0) taskDesc += "⌛";
            if(task["Status"] == 1) taskDesc += task["WorkerMsg"] + " 👷";
            if(task["Status"] == 2) taskDesc += task["Output"] + " ✅";
            if(task["Status"] == 3) taskDesc += "⛔";
            let lblRef = this.AddElem(
              divTask, 'span', {'innerHTML':taskDesc});
            // Delete button
            let btnDelete = this.AddElem(divTask, 'input',
              {'type':'button', 'value': 'delete'});
            btnDelete.addEventListener('click', (event) => {
              taskQueue.DeleteTask(ref, task["Label"]);});
            // Memorise the task
            this.tasks[ref] = task;
          });
        }
      });
  }
  // Delete a task
  DeleteTask(refTask, labelTask) {
    // Create the API request data
    let data = {
      'action': 'deleteTask',
      'ref': refTask,
      'owner': this.login,
      'label': labelTask,
    };
    let reqData = this.DictToReqData(data);
    // Post the API request
    this.PostRequest(reqData,
      // Handler for the API request response
      (retData) => {
        // If the deleted task is the displayed task, undisplay it
        if(this.currentTask == refTask) this.ViewTask(0);
        // Update the DOM
        document.getElementById("divTask" + refTask).remove();
        this.tasks[refTask] = {};
      });
  }
}
// Variable to memorise the global TaskQueue instance
var taskQueue = {};
// Handler for the onload event
window.onload = function() {
  taskQueue = new TaskQueue();
  if(TaskQueue.login != "") {
    // Set all the handlers in the DOM
    elem('btnAddTask').addEventListener('click', (event) => {
      taskQueue.AddTask();});
    // Initialise the task list
    taskQueue.UpdateTasks();
    // Start the automatic refresh of the task list every 1 second
    setInterval(function() {
      taskQueue.UpdateTasks();
     }, 1000);
  }
}
  </script> 
</html>
